@@ -83,15 +83,34 @@ write_position(uint8_t *buf, const struct sc_position *position) {
     sc_write16be(&buf[10], position->screen_size.height);
 }
 
-// write length (4 bytes) + string (non null-terminated)
+// Write truncated string, and return the size
 static size_t
-write_string(const char *utf8, size_t max_len, uint8_t *buf) {
+write_string_payload(uint8_t *payload, const char *utf8, size_t max_len) {
+    if (!utf8) {
+        return 0;
+    }
     size_t len = sc_str_utf8_truncation_index(utf8, max_len);
+    memcpy(payload, utf8, len);
+    return len;
+}
+
+// Write length (4 bytes) + string (non null-terminated)
+static size_t
+write_string(uint8_t *buf, const char *utf8, size_t max_len) {
+    size_t len = write_string_payload(buf + 4, utf8, max_len);
     sc_write32be(buf, len);
-    memcpy(&buf[4], utf8, len);
     return 4 + len;
 }
 
+// Write length (1 byte) + string (non null-terminated)
+static size_t
+write_string_tiny(uint8_t *buf, const char *utf8, size_t max_len) {
+    assert(max_len <= 0xFF);
+    size_t len = write_string_payload(buf + 1, utf8, max_len);
+    buf[0] = len;
+    return 1 + len;
+}
+
 size_t
 sc_control_msg_serialize(const struct sc_control_msg *msg, uint8_t *buf) {
     buf[0] = msg->type;
@@ -103,9 +122,8 @@ sc_control_msg_serialize(const struct sc_control_msg *msg, uint8_t *buf) {
             sc_write32be(&buf[10], msg->inject_keycode.metastate);
             return 14;
         case SC_CONTROL_MSG_TYPE_INJECT_TEXT: {
-            size_t len =
-                write_string(msg->inject_text.text,
-                             SC_CONTROL_MSG_INJECT_TEXT_MAX_LENGTH, &buf[1]);
+            size_t len = write_string(&buf[1], msg->inject_text.text,
+                                      SC_CONTROL_MSG_INJECT_TEXT_MAX_LENGTH);
             return 1 + len;
         }
         case SC_CONTROL_MSG_TYPE_INJECT_TOUCH_EVENT:
@@ -137,24 +155,34 @@ sc_control_msg_serialize(const struct sc_control_msg *msg, uint8_t *buf) {
         case SC_CONTROL_MSG_TYPE_SET_CLIPBOARD:
             sc_write64be(&buf[1], msg->set_clipboard.sequence);
             buf[9] = !!msg->set_clipboard.paste;
-            size_t len = write_string(msg->set_clipboard.text,
-                                      SC_CONTROL_MSG_CLIPBOARD_TEXT_MAX_LENGTH,
-                                      &buf[10]);
+            size_t len = write_string(&buf[10], msg->set_clipboard.text,
+                                      SC_CONTROL_MSG_CLIPBOARD_TEXT_MAX_LENGTH);
             return 10 + len;
         case SC_CONTROL_MSG_TYPE_SET_SCREEN_POWER_MODE:
             buf[1] = msg->set_screen_power_mode.mode;
             return 2;
         case SC_CONTROL_MSG_TYPE_UHID_CREATE:
             sc_write16be(&buf[1], msg->uhid_create.id);
-            sc_write16be(&buf[3], msg->uhid_create.report_desc_size);
-            memcpy(&buf[5], msg->uhid_create.report_desc,
-                            msg->uhid_create.report_desc_size);
-            return 5 + msg->uhid_create.report_desc_size;
+
+            size_t index = 3;
+            index += write_string_tiny(&buf[index], msg->uhid_create.name, 127);
+
+            sc_write16be(&buf[index], msg->uhid_create.report_desc_size);
+            index += 2;
+
+            memcpy(&buf[index], msg->uhid_create.report_desc,
+                                msg->uhid_create.report_desc_size);
+            index += msg->uhid_create.report_desc_size;
+
+            return index;
         case SC_CONTROL_MSG_TYPE_UHID_INPUT:
             sc_write16be(&buf[1], msg->uhid_input.id);
             sc_write16be(&buf[3], msg->uhid_input.size);
             memcpy(&buf[5], msg->uhid_input.data, msg->uhid_input.size);
             return 5 + msg->uhid_input.size;
+        case SC_CONTROL_MSG_TYPE_UHID_DESTROY:
+            sc_write16be(&buf[1], msg->uhid_destroy.id);
+            return 3;
         case SC_CONTROL_MSG_TYPE_EXPAND_NOTIFICATION_PANEL:
         case SC_CONTROL_MSG_TYPE_EXPAND_SETTINGS_PANEL:
         case SC_CONTROL_MSG_TYPE_COLLAPSE_PANELS:
@@ -252,10 +280,15 @@ sc_control_msg_log(const struct sc_control_msg *msg) {
         case SC_CONTROL_MSG_TYPE_ROTATE_DEVICE:
             LOG_CMSG("rotate device");
             break;
-        case SC_CONTROL_MSG_TYPE_UHID_CREATE:
-            LOG_CMSG("UHID create [%" PRIu16 "] report_desc_size=%" PRIu16,
-                     msg->uhid_create.id, msg->uhid_create.report_desc_size);
+        case SC_CONTROL_MSG_TYPE_UHID_CREATE: {
+            // Quote only if name is not null
+            const char *name = msg->uhid_create.name;
+            const char *quote = name ? "\"" : "";
+            LOG_CMSG("UHID create [%" PRIu16 "] name=%s%s%s "
+                     "report_desc_size=%" PRIu16, msg->uhid_create.id,
+                     quote, name, quote, msg->uhid_create.report_desc_size);
             break;
+        }
         case SC_CONTROL_MSG_TYPE_UHID_INPUT: {
             char *hex = sc_str_to_hex_string(msg->uhid_input.data,
                                              msg->uhid_input.size);
@@ -269,6 +302,9 @@ sc_control_msg_log(const struct sc_control_msg *msg) {
             }
             break;
         }
+        case SC_CONTROL_MSG_TYPE_UHID_DESTROY:
+            LOG_CMSG("UHID destroy [%" PRIu16 "]", msg->uhid_destroy.id);
+            break;
         case SC_CONTROL_MSG_TYPE_OPEN_HARD_KEYBOARD_SETTINGS:
             LOG_CMSG("open hard keyboard settings");
             break;
@@ -278,6 +314,16 @@ sc_control_msg_log(const struct sc_control_msg *msg) {
     }
 }
 
+bool
+sc_control_msg_is_droppable(const struct sc_control_msg *msg) {
+    // Cannot drop UHID_CREATE messages, because it would cause all further
+    // UHID_INPUT messages for this device to be invalid.
+    // Cannot drop UHID_DESTROY messages either, because a further UHID_CREATE
+    // with the same id may fail.
+    return msg->type != SC_CONTROL_MSG_TYPE_UHID_CREATE
+        && msg->type != SC_CONTROL_MSG_TYPE_UHID_DESTROY;
+}
+
 void
 sc_control_msg_destroy(struct sc_control_msg *msg) {
     switch (msg->type) {
