@@ -29,7 +29,7 @@ public class Options {
     private boolean audioDup;
     private int videoBitRate = 8000000;
     private int audioBitRate = 128000;
-    private int maxFps;
+    private float maxFps;
     private int lockVideoOrientation = -1;
     private boolean tunnelForward;
     private Rect crop;
@@ -113,7 +113,7 @@ public class Options {
         return audioBitRate;
     }
 
-    public int getMaxFps() {
+    public float getMaxFps() {
         return maxFps;
     }
 
@@ -321,7 +321,7 @@ public class Options {
                     options.audioBitRate = Integer.parseInt(value);
                     break;
                 case "max_fps":
-                    options.maxFps = Integer.parseInt(value);
+                    options.maxFps = parseFloat("max_fps", value);
                     break;
                 case "lock_video_orientation":
                     options.lockVideoOrientation = Integer.parseInt(value);
@@ -456,8 +456,14 @@ public class Options {
         }
         int width = Integer.parseInt(tokens[0]);
         int height = Integer.parseInt(tokens[1]);
+        if (width <= 0 || height <= 0) {
+            throw new IllegalArgumentException("Invalid crop size: " + width + "x" + height);
+        }
         int x = Integer.parseInt(tokens[2]);
         int y = Integer.parseInt(tokens[3]);
+        if (x < 0 || y < 0) {
+            throw new IllegalArgumentException("Invalid crop offset: " + x + ":" + y);
+        }
         return new Rect(x, y, x + width, y + height);
     }
 
@@ -487,4 +493,12 @@ public class Options {
         float floatAr = Float.parseFloat(tokens[0]);
         return CameraAspectRatio.fromFloat(floatAr);
     }
+
+    private static float parseFloat(String key, String value) {
+        try {
+            return Float.parseFloat(value);
+        } catch (NumberFormatException e) {
+            throw new IllegalArgumentException("Invalid float value for " + key + ": \"" + value + "\"");
+        }
+    }
 }
